
// includes
#include <fstream>
#include <sstream>

#include "../lang/Lexer.hpp"
#include "../lang/Parser.hpp"

#include "mARBle.hpp"

using namespace bleifrei::lang;
using namespace std;

namespace bleifrei {
	namespace gfx {
		namespace mARBle {

			map<string, string> defines;

			void clearDefines(void)
			{
				defines.clear();
			}

			void setDefine(const std::string &identifier, const std::string &def)
			{
				defines.insert(make_pair(identifier, def));
			}

			enum MarbleKeyword {
				Define,
				Include,
				Ifdef,
				Ifndef,
				Else,
				Endif,
				MNRM,
				MM4x4,
				MM3x3,
				END
			};

			class ConvertLexer : public Lexer {
			public:
				ConvertLexer(istream &is) : Lexer(is, "", ",;")
				{
					setCStrings(false);
					setSkipLinebreak(false);
					setCaseSensitive(true);
					addKeyword("define", Define);
					addKeyword("include", Include);
					addKeyword("ifdef", Ifdef);
					addKeyword("ifndef", Ifndef);
					addKeyword("else", Else);
					addKeyword("endif", Endif);
					addKeyword("mNRM", MNRM);
					addKeyword("mM4X4", MM4x4);
					addKeyword("mM3X3", MM3x3);
					addKeyword("END", END);
				}
			};

			class ConvertParser : public Parser {
			public:
				ConvertParser(ConvertLexer &lex) : Parser(lex) {}
				template<class T> bool test(T &token)	{ return Parser::test(token);	}
				template<class T> void parse(T &token)	{ Parser::parse(token);			}
			};

			template<class T>
			inline Matcher<ConvertParser> operator <<(ConvertParser &parser, T &type)
			{
				if (parser.test(type))
					return Matcher<ConvertParser>(parser);
				else
					return Matcher<ConvertParser>(parser, false);
			}

			template<class T>
			inline Matcher<ConvertParser> &operator <<(Matcher<ConvertParser> &matcher, T &type)
			{
				if (matcher.valid) matcher.parser.parse(type);
				return matcher;
			}

			inline const string &replace(const string &identifier, map<string, string> &defines)
			{
				map<string, string>::iterator it = defines.find(identifier);

				if (it == defines.end()) return identifier;
				
				return it->second;
			}

			inline void getLine(ConvertLexer &lex, string &str)
			{
				char ch;
				while (lex.getChar(ch) && ch != '\n') str += ch;
			}

			bool convert(istream &is, ostream &os, bool clear_defines)
			{
				ConvertLexer lex(is);
				ConvertParser parser(lex);

				bool inside_ifdef = false;

				if (clear_defines) clearDefines();

				try {
					while (!is.eof()) {
						string define, include;
						Identifier dst, src, mat, vec, def;

						lex.pushState();
						if (parser << Keyword("define") << def << define << Sep(";")) {
							setDefine(def.str, define);
						}
						else if (parser << Keyword("include") << include << Sep(";")) {
							convert(ifstream(include.c_str()), os, false);
						}
						else if (parser << Keyword("ifndef") << def) {
							if (inside_ifdef) throw expected("endif", "ifndef", lex);
							inside_ifdef = true;

							if (defines.find(def.str) != defines.end()) {
								while (!is.eof()) {
									if (parser << Keyword("else")) {
										inside_ifdef = true;
										break;
									}
									else if (parser << Keyword("endif")) {
										inside_ifdef = false;
										break;
									}
									lex.flush();
									string line;
									getLine(lex, line);
								}
							}
						}
						else if (parser << Keyword("ifdef") << def) {
							if (inside_ifdef) throw expected("endif", "ifdef", lex);
							inside_ifdef = true;

							if (defines.find(def.str) == defines.end()) {
								while (!is.eof()) {
									if (parser << Keyword("else")) {
										inside_ifdef = true;
										break;
									}
									else if (parser << Keyword("endif")) {
										inside_ifdef = false;
										break;
									}
									lex.flush();
									string line;
									getLine(lex, line);
								}
							}
						}
						else if (parser << Keyword("else")) {
							if (!inside_ifdef) throw expected("statement", "else without ifdef", lex);
							
							while (!(parser << Keyword("endif")) && !is.eof()) {
								lex.flush();
								string line;
								getLine(lex, line);
							}
							inside_ifdef = false;
						}
						else if (parser << Keyword("endif")) {
							if (!inside_ifdef) throw expected("statement", "endif without ifdef", lex);
							inside_ifdef = false;
						}
						else if (parser << Keyword("mM4X4") << dst << Sep(",") << mat << Sep(",") << vec << Sep(";")) {
							string line;
							getLine(lex, line);

							const string &d = replace(dst.str, defines);
							const string &m = replace(mat.str, defines);
							const string &v = replace(vec.str, defines);
							
							os << "DP4 " << d << ".x, " << m << "[0], " << v << ";" << line << endl;
							os << "DP4 " << d << ".y, " << m << "[1], " << v << ";" << endl;
							os << "DP4 " << d << ".z, " << m << "[2], " << v << ";" << endl;
							os << "DP4 " << d << ".w, " << m << "[3], " << v << ";" << endl;
						}
						else if (parser << Keyword("mM3X3") << dst << Sep(",") << mat << Sep(",") << vec << Sep(";")) {
							string line;
							getLine(lex, line);

							const string &d = replace(dst.str, defines);
							const string &m = replace(mat.str, defines);
							const string &v = replace(vec.str, defines);
											
							os << "DP3 " << d << ".x, " << m << "[0], " << v << ";" << line << endl;
							os << "DP3 " << d << ".y, " << m << "[1], " << v << ";" << endl;
							os << "DP3 " << d << ".z, " << m << "[2], " << v << ";" << endl;
						}
						else if (parser << Keyword("mNRM") << dst << Sep(",") << src << Sep(";")) {
							string line;
							getLine(lex, line);

							const string &d = replace(dst.str, defines);
							const string &s = replace(src.str, defines);
							
							os << "DP3 " << d << ".w, " << s << ", " << s << ";" << line << endl;
							os << "RSQ " << d << ".w, " << d << ".w;" << endl;
							os << "MUL " << d << ".xyz, " << s << ", " << d << ".w;" << endl;
						}
						else if (parser << Keyword("END")) {
							string line;
							getLine(lex, line);

							os << "END" << line << endl;

							break;
						}
						else {
							lex.flush();
							lex.popState();

							char ch;
							lex.getChar(ch);
							while (!is.eof() && ch != '\n') {
								while (!isalnum(ch) && ch != '_' && ch != '\n') {
									os.put(ch);
									if (ch == '#') {
										string line;
										getLine(lex, line);
										os << line << endl;
										ch = '\n';
									}
									else {
										lex.getChar(ch);
									}
								}

								string ident;
								if (isalpha(ch) || ch == '_') {
									ident += ch;
									lex.getChar(ch);
									while (isalnum(ch) || ch == '_') {
										ident += ch;
										lex.getChar(ch);
									}
								}
								else if (ch != '\n') {
									os.put(ch);
									while (lex.getChar(ch) && isalpha(ch)) os.put(ch);
								}

								os << replace(ident, defines);
							}
							os << endl;
						}
					}
				}
				catch (expected &e) {
					cerr << "Error on line " << e.line << ", column " << e.col << ":" << endl;
					cerr << "\tExpected " << e.expect << ", found " << e.found << endl;
				}
				catch (Lexer::newline_in_constant) {
					cerr << "Error on line " << lex.getCurrentLine() << ", column " << lex.getCurrentColumn() << ":" << endl;
					cerr << "\tNewline in constant" << endl;
				}

				return false;
			} 

			FragmentProgram loadFragmentProgram(std::istream &is, bool clear_defines)
			{
				stringstream temp;
				convert(is, temp, clear_defines);
				return FragmentProgram(temp);
			}

			VertexProgram loadVertexProgram(std::istream &is, bool clear_defines)
			{
				stringstream temp;
				convert(is, temp, clear_defines);
				return VertexProgram(temp);
			}

		}
	}
}